include("scripts/Cam/CamExport/CamExporter.js");

function LCNC(documentInterface, newDocumentInterface) {
    CamExporter.call(this, documentInterface, newDocumentInterface);

    this.toolPosition = LCNC.ToolPosition.Up;
    this.alwaysWriteGCode = false;
    this.g = undefined;
    this.gPrev = undefined;
    this.x = undefined;
    this.xPrev = undefined;
    this.y = undefined;
    this.yPrev = undefined;
    this.i = undefined;
    this.j = undefined;
    this.z = undefined;
    this.zPrev = undefined;
    this.f = undefined;
    this.fPrev = undefined;

    this.separator = " ";

    this.globalOptions = "GCode";
    this.layerOptions = "GCodeLayer";
}

LCNC.prototype = new CamExporter();

LCNC.Mode = {
    Rapid : 0,
    Normal : 1,
    CircularCW: 2,
    CircularCCW: 3
};

LCNC.ToolPosition = {
    Clear : 0,
    Up : 1,
    Down : 2
};

LCNC.prototype.initGlobalOptionWidget = function(w) {
    switch (w.objectName) {
    case "ZSafety":
        w.addItems(["50", "30", "20", "10"]);
        w.setEditText("20");
        break;
    case "ZClear":
        w.addItems(["1", "2", "3"]);
        w.setEditText("2");
        break;
    case "ZCutting":
        w.addItems(["-0.5", "-1", "-1.5", "-2"]);
        w.setEditText("-1");
        break;
    }
};

LCNC.prototype.initLayerOptionWidget = function(w) {
    switch (w.objectName) {
    case "ZCutting":
        w.addItems(["default", "-1", "-2", "-3"]);
        w.currentText = "default";
        break;
    }
};

LCNC.prototype.getFileExtensions = function() {
    return ["ngc"];
};

LCNC.prototype.getRapidMoveCode = function() {
    return "G00";
};

LCNC.prototype.getLinearMoveCode = function() {
    return "G01";
};

LCNC.prototype.getCircularCWMoveCode = function() {
    return "G02";
};

LCNC.prototype.getCircularCCWMoveCode = function() {
    return "G03";
};


LCNC.prototype.getXCode = function(value) {
    var v = value;
    if (this.getIncrementalXYZ()) {
        v -= this.xPrev;
    }
    return sprintf("X%.3f", v);
};

LCNC.prototype.getYCode = function(value) {
    var v = value;
    if (this.getIncrementalXYZ()) {
        v -= this.yPrev;
    }
    return sprintf("Y%.3f", v);
};

LCNC.prototype.getZCode = function(value) {
    var v = value;
    if (this.getIncrementalXYZ()) {
        v -= this.zPrev;
    }
    return sprintf("Z%.3f", v);
};

LCNC.prototype.getICode = function(value) {
    return sprintf("I%.3f", value);
};

LCNC.prototype.getJCode = function(value) {
    return sprintf("J%.3f", value);
};

LCNC.prototype.getFCode = function(value) {
    return sprintf("F%d", value);
};


LCNC.prototype.getSafetyZLevel = function() {
    return parseFloat(this.document.getVariable("Cam/ZSafety", 20.0));
};

LCNC.prototype.getToolUpLevel = function() {
    return parseFloat(this.document.getVariable("Cam/ZClear", 2.0));
};

LCNC.prototype.getToolDownLevel = function() {
    // document wide default value:
    var docValue = parseFloat(this.document.getVariable("Cam/ZCutting", -1.0));

    var entity = this.getEntity();
    if (isNull(entity)) {
        return docValue;
    }

    // layer specific value:
    var layerId = entity.getLayerId();
    var layer = this.document.queryLayer(layerId);
    if (isNull(layer)) {
        return docValue;
    }

    var layerValue = layer.getCustomProperty("QCADCAM", "Cam/ZCutting", docValue);
    if (layerValue==="default") {
        return docValue;
    }
    else {
        return layerValue;
    }
};


LCNC.prototype.getFeedrate = function() {
    return parseFloat(this.document.getVariable("Cam/Feedrate", 150));
};



LCNC.prototype.toolIsUp = function() {
    this.g = LCNC.Mode.Rapid;
    this.toolPosition = LCNC.ToolPosition.Up;
};

LCNC.prototype.toolIsDown = function() {
    this.g = LCNC.Mode.Normal;
    this.toolPosition = LCNC.ToolPosition.Down;
};



LCNC.prototype.writeToolUp = function() {
    this.g = LCNC.Mode.Rapid;
    this.z = this.getToolUpLevel();
    this.toolPosition = LCNC.ToolPosition.Up;
    if (this.feedRateSet!==true) {
        this.writeLine(undefined, "F150");
        this.feedRateSet=true;
    }
    else {
        this.writeLine();
    }
    this.toolIsUp();
};

LCNC.prototype.writeToolDown = function() {
    this.g = LCNC.Mode.Normal;
    this.z = this.getToolDownLevel();
    this.toolPosition = LCNC.ToolPosition.Down;
    if (this.feedRateSet!==true) 
        {
        this.writeLine(undefined, "F150");
        this.feedRateSet=true;
        }
        else 
        {
        this.writeLine();
        }
    this.toolIsDown();
};

LCNC.prototype.writeRapidZMove = function(z) {
    this.g = LCNC.Mode.Rapid;
    this.z = z;
    this.writeLine();
};

LCNC.prototype.writeZMove = function(z) {
    this.g = LCNC.Mode.Normal;
    this.z = z;
    this.writeLine();
};

// G00:
LCNC.prototype.writeRapidLinearMove = function(x, y) {
    if (!this.gotXMove(x) && !this.gotYMove(y)) {
        return;
    }

    this.writeBeforeRapidLinearMove(x, y);

    // force tool up before Rapid move:
    if (this.toolPosition !== LCNC.ToolPosition.Up && this.toolPosition !== LCNC.ToolPosition.Clear) {
        this.writeToolUp();
    }

    this.g = LCNC.Mode.Rapid;
    this.x = x;
    this.y = y;
    this.writeLine();

    this.writeAfterRapidLinearMove(x, y);
};

LCNC.prototype.writeBeforeRapidLinearMove = function(x, y) { };
LCNC.prototype.writeAfterRapidLinearMove = function(x, y) { };

LCNC.prototype.prepareForCutting = function() {
    // force tool down before normal move:
    if (this.toolPosition !== LCNC.ToolPosition.Down) {
        // force tool up before tool down:
        if (this.toolPosition !== LCNC.ToolPosition.Up) {
            this.writeToolUp();
        }
        this.writeToolDown();
    }
};

// G01:
LCNC.prototype.writeLinearMove = function(x, y) {
    this.prepareForCutting();

    this.g = LCNC.Mode.Normal;
    this.x = x;
    this.y = y;
    this.writeLine();
};

// G02 / G03:
LCNC.prototype.writeCircularMove = function(x, y,
    center, radius,
    startAngle, endAngle,
    isLarge, isReversed) {

    this.prepareForCutting();

    if (isReversed) {
        this.g = LCNC.Mode.CircularCW;
    }
    else {
        this.g = LCNC.Mode.CircularCCW;
    }

    this.i = center.x - this.x;
    this.j = center.y - this.y;
    this.x = x;
    this.y = y;
    this.writeLine();
};

LCNC.prototype.writeHeader = function() {
    this.writeLine("G21 G17 G40 G49 G54 G64 P0.1 G80 G90 G94");
    this.writeRapidZMove(this.getSafetyZLevel());
    this.toolPosition = LCNC.ToolPosition.Clear;
};

LCNC.prototype.writeFooter = function() {
    this.writeToolUp();
    this.writeRapidZMove(this.getSafetyZLevel());
    this.toolPosition = LCNC.ToolPosition.Clear;
    this.writeLine("M2");
};

LCNC.prototype.gotModeChange = function(m) {
    return (isNull(this.gPrev) && !isNull(m)) || (!isNull(m) && this.gPrev!=m);
};

LCNC.prototype.gotXMove = function(x) {
    return (isNull(this.xPrev) && !isNull(x)) || (!isNull(x) && !this.fuzzyCompare(x, this.xPrev));
};

LCNC.prototype.gotYMove = function(y) {
    return (isNull(this.yPrev) && !isNull(y)) || (!isNull(y) && !this.fuzzyCompare(y, this.yPrev));
};

LCNC.prototype.gotZMove = function(z) {
    return (isNull(this.zPrev) && !isNull(z)) || (!isNull(z) && !this.fuzzyCompare(z, this.zPrev));
};

LCNC.prototype.gotZUpMove = function(z) {
    return this.gotZMove(z) && z>this.zPrev;
};

LCNC.prototype.gotFeedrateChange = function(f) {
    return (isNull(this.fPrev) && !isNull(f)) || (!isNull(f) && f!==this.fPrev);
};

LCNC.prototype.getCamLayerId = function() {
    if (this.duringRapidMove===true) {
        return CamExporter.prototype.getCamLayerId.call(this);
    }

    return this.createLayer("cut at " + this.getToolDownLevel(), new RColor("white"));
};

LCNC.prototype.append = function(line, str) {
    if (line.length===0) {
        return str;
    }
    else {
        return line + this.separator + str;
    }
};

/**
 * Writes the next line of the file or the given custom line.
 *
 * \param custom string (optional) custom line contents
 * \param append string (optional) append to line contents
 */
LCNC.prototype.writeLine = function(custom, append) {
    var line = "";

    if (!isNull(custom)) {
        line = custom;
        CamExporter.prototype.writeLine.call(this, line);
        return;
    }

    var gotModeChange = this.gotModeChange(this.g);
    var gotXMove = this.gotXMove(this.x);
    var gotYMove = this.gotYMove(this.y);
    var gotZMove = this.gotZMove(this.z);
    var gotFeedrateChange = this.gotFeedrateChange(this.f);

    // nothing to do:
    if (this.g!==LCNC.Mode.CircularCW && this.g!==LCNC.Mode.CircularCCW &&
        !gotXMove && !gotYMove && !gotZMove && !gotFeedrateChange) {
        return;
    }

    switch (this.g) {
    case LCNC.Mode.Rapid:
        if (gotModeChange) {
            line = this.append(line, this.getRapidMoveCode());
        }
        break;
    case LCNC.Mode.Normal:
        if (gotModeChange) {
            line = this.append(line, this.getLinearMoveCode());
        }
        break;
    case LCNC.Mode.CircularCW:
        line = this.append(line, this.getCircularCWMoveCode());
        break;
    case LCNC.Mode.CircularCCW:
        line = this.append(line, this.getCircularCCWMoveCode());
        break;
    }

    if (gotXMove || this.g===LCNC.Mode.CircularCW || this.g===LCNC.Mode.CircularCCW) {
        line = this.append(line, this.getXCode(this.x));
    }

    if (gotYMove || this.g===LCNC.Mode.CircularCW || this.g===LCNC.Mode.CircularCCW) {
        line = this.append(line, this.getYCode(this.y));
    }

    if (gotZMove) {
        line = this.append(line, this.getZCode(this.z));
    }

    if (gotFeedrateChange) {
        line = this.append(line, this.getFCode(this.f));
    }

    if (this.g===LCNC.Mode.CircularCW || this.g===LCNC.Mode.CircularCCW) {
        line = this.append(line, this.getICode(this.i));
        line = this.append(line, this.getJCode(this.j));
    }

    if (!isNull(append)) {
        line = this.append(line, append);
    }

    CamExporter.prototype.writeLine.call(this, line);

    this.xPrev = this.x;
    this.yPrev = this.y;
    this.zPrev = this.z;
    this.fPrev = this.f;
    this.gPrev = this.g;
};
